#include <stdlib.h>
#include <stdio.h>
#include <cutil.h>
#include <cuda.h>
#include <kernel.cu>


int getCudaDevice()
{
	int deviceCount = -1;
	cudaGetDeviceCount(&deviceCount);	
	if(deviceCount < 1)
	{
		printf("Could not find any devices\n");	
		return 1;
	}
	else
	{
		printf("CUDA device count: %d\n", deviceCount);
		cudaSetDevice(0);
		return 0;
	}	
}

extern "C" void run_test(float* data, int len)
{
	if(0 != getCudaDevice())
		return;
		
	//allocate device memory and copy data from host
	const unsigned int mem_size = len*sizeof(float);
	float* d_data;
	cudaMalloc((void**)&d_data, mem_size);
	cudaMemcpy(d_data, data, mem_size, cudaMemcpyHostToDevice);

	dim3 grid(1,1,1);
	dim3 threads(len, 1, 1);
	kernel<<<grid,threads>>>(d_data);

	cudaMemcpy(data, d_data, mem_size, cudaMemcpyDeviceToHost);

	cudaFree(d_data);
}

extern "C" float run_pingpong(int* data_slow, int numElements, int timeToRunInMs)
{
        //Get a CUDA device
	if(0 != getCudaDevice())
		return -1.0f;
	
        //Data pointers
        int *h_data, *d_data;                   
        
        //Calculate memory size to allocate
        const unsigned int mem_size = numElements*sizeof(int);                
        
        //Alloc faster page-locked host memory
        cudaMallocHost((void**)&h_data, mem_size);
        
        //h_data = (int*)malloc(mem_size);
        
        //Alloc device memory
        cudaMalloc((void**)&d_data, mem_size);        
                
        //Copy data from slower pageable memory to page-locked memory
        for(int i = 0; i < numElements; ++i)
        {
            h_data[i] = data_slow[i];
        }
               
        //Copy data from host to device
        //cudaMemcpy(d_data, h_data, mem_size, cudaMemcpyHostToDevice);
        
        //Set up threads and run kernel
        //dim3 grid(1,1,1);
        //dim3 threads(numElements, 1, 1);    
        //pingpong_kernel<<<grid,threads>>>(d_data);
        
        //Create a timer
        unsigned int timer = 0;
        float elapsedTimeInMs = 0.0f;
        cutCreateTimer(&timer);
        
        //Iteration counter
        unsigned int numIterations = 0;
        
	//Copy data from Host to Device then back again for a user defined time duration
        while(elapsedTimeInMs < timeToRunInMs)
        {
            //Start CUDA timer
            cutStartTimer(timer);
                        
            //Copy data from host to device
            cudaMemcpy(d_data, h_data, mem_size, cudaMemcpyHostToDevice);
            
            //pingpong_kernel<<<grid,threads>>>(d_data);

            //Copy data from device to host
            cudaMemcpy(h_data, d_data, mem_size, cudaMemcpyDeviceToHost);
            
            //Increment number of iterations
            ++numIterations;
            
            //Stop timer and get current time used
            cutStopTimer(timer);  
            elapsedTimeInMs = cutGetTimerValue(timer);            
        }
        
        //Set host data to 1
        for(int i = 0; i < numElements; ++i)
        {
            h_data[i] = 1;
        }
        
        //Copy new data from host to device
        cudaMemcpy(d_data, h_data, mem_size, cudaMemcpyHostToDevice);
        
        //Wait for kernel update
        unsigned int counter = 0;
        while(++counter != 1000000){}
        
        //Copy kernel modified data from device to host
        cudaMemcpy(h_data, d_data, mem_size, cudaMemcpyDeviceToHost);

        //Copy data from local host memory to global for print in main
        for(int i = 0; i < numElements; ++i)
        {
            data_slow[i] = h_data[i];
        }
        
        //Free data
        cudaFree(d_data);
        cudaFreeHost(h_data);
        
        //Return number of iterations pr. second
        return (float) numIterations/(elapsedTimeInMs/1000);        
}

extern "C" void run_matMul(float* h_a, float* h_b, float* h_c, int wa, int ha, int wb, int hb)
{
		if(0 != getCudaDevice())
		return;
		
#ifdef DEBUG
		printf("A length: %d\n", wa*ha);
		printf("B length: %d\n", wb*hb);
		printf("C length: %d\n", ha*wb);	
#endif
					
		//Calc size of matrices
		const unsigned int mem_size_a = wa*ha*sizeof(float);
		const unsigned int mem_size_b = wb*hb*sizeof(float);
		const unsigned int mem_size_c = ha*wb*sizeof(float);
		float* d_a;
		float* d_b;
		float* d_c;
		
		//Alloc device memory
		cudaMalloc((void**)&d_a, mem_size_a);
		cudaMalloc((void**)&d_b, mem_size_b);
		cudaMalloc((void**)&d_c, mem_size_c);
		cudaMemcpy(d_a, h_a, mem_size_a, cudaMemcpyHostToDevice);
		cudaMemcpy(d_b, h_b, mem_size_b, cudaMemcpyHostToDevice);
		cudaMemcpy(d_c, h_c, mem_size_c, cudaMemcpyHostToDevice);
		
		//Define grid and thread size
		dim3 grid(ha, 1, 1);
		dim3 threads(wb, 1, 1);
		
		//run matrix multiply on device
		matMul_kernel<<<grid,threads>>>(d_a, d_b, d_c, wa, ha, wb, hb);
		
		cudaMemcpy(h_c, d_c, mem_size_c, cudaMemcpyDeviceToHost);
		
		cudaFree(d_a);
		cudaFree(d_b);
		cudaFree(d_c);
}
