#include "CTask.h"
#include "ThreadHandler.h"
#include "CThread.h"
#include "TaskQueue.h"

//--------------------------------------------------------------------------------------------
// Default constructor 
//--------------------------------------------------------------------------------------------
CThread::CThread() :
    _hasInitThreadAttribute(false),
    _suspend(true),
    _exit(false),
    _task(0),
    _id(-1),
    _threadHandler(0)
{
    if(false == _hasInitThreadAttribute)
    {
        initThreadAttribute();
    }
}

//--------------------------------------------------------------------------------------------
// Copy constructor
//--------------------------------------------------------------------------------------------
CThread::CThread(const CThread & in) :
    _thread(in._thread),
    _thread_attribute(in._thread_attribute),
    _mutex_suspend(in._mutex_suspend),
    _cond_resume(in._cond_resume),
    _hasInitThreadAttribute(in._hasInitThreadAttribute),
    _suspend(in._suspend),
    _exit(in._exit),
    _task(in._task),
    _id(in._id),
    _threadHandler(in._threadHandler)
{
    if(false == _hasInitThreadAttribute)
    {
        initThreadAttribute();
    }
}

//--------------------------------------------------------------------------------------------
// Copy contstructor
//--------------------------------------------------------------------------------------------
CThread &CThread::operator=(const CThread & in) 
{    
    if(this == &in) return *this;
    _thread = in._thread;
    _thread_attribute = in._thread_attribute;
    _mutex_suspend = in._mutex_suspend;
    _cond_resume = in._cond_resume;
    _hasInitThreadAttribute = in._hasInitThreadAttribute;
    _suspend = in._suspend;
    _exit = in._exit;
    _task = in._task;
    _id = in._id;
    _threadHandler = in._threadHandler;

    if(false == _hasInitThreadAttribute)
    {
        initThreadAttribute();
    }
    return *this;
}

//--------------------------------------------------------------------------------------------
// Destructor
//--------------------------------------------------------------------------------------------
CThread::~CThread() 
{
    //Destroy thread attributes
    if(true == _hasInitThreadAttribute)
    {
        pthread_attr_destroy(&_thread_attribute);
        pthread_mutex_destroy(&_mutex_suspend);
        pthread_mutex_destroy(&_mutex_state);
        pthread_cond_destroy(&_cond_resume);
    }

//    //Delete task
//    if(0 != _task)
//    {
//        delete _task;
//        _task = 0;
//    }

    //Exit thread
//    pthread_exit(0);
}

//--------------------------------------------------------------------------------------------
// Initializes the thread attributes
//--------------------------------------------------------------------------------------------
void CThread::initThreadAttribute()
{ 
    if(true == _hasInitThreadAttribute)
        return;

    pthread_attr_init(&_thread_attribute); 
    pthread_attr_setdetachstate(&_thread_attribute, PTHREAD_CREATE_JOINABLE); 
    pthread_mutex_init(&_mutex_suspend, NULL);
    pthread_mutex_init(&_mutex_state, NULL);
    pthread_cond_init(&_cond_resume, NULL);
    _hasInitThreadAttribute = true;
}   

//--------------------------------------------------------------------------------------------
// Creates the thread with itself as input 
//--------------------------------------------------------------------------------------------
void CThread::create()
{
    int retval = pthread_create(&_thread, &_thread_attribute, theWork, (void*) this );
    if(0 != retval)
    {
        printf("CThread.cpp: Pthread create retval: %d\n", retval);
    }
}

//--------------------------------------------------------------------------------------------
// Resumes the thread with new task
//--------------------------------------------------------------------------------------------
bool CThread::resume(CTask* task)
{
    //Lock suspend mutex to make sure worker is sleeping before signal
    pthread_mutex_lock(&_mutex_suspend);

    //Set thread task
    _task = task;

#ifdef DEBUG
    printf("CThread %d: Signaling resume with task %d\n", _id, _task->getId());        
#endif

    //Signal thread to wake up
    int retVal = pthread_cond_signal(&_cond_resume);
    if(0 != retVal)
    {
        printf("CThread.cpp: Error resuming thread. Error: %d\n", retVal);
        return false;
    }

    //Unlock suspend mutex so signaled thread can regain it
    pthread_mutex_unlock(&_mutex_suspend);
    return true;
}

//--------------------------------------------------------------------------------------------
// Checks if thread is suspended
//--------------------------------------------------------------------------------------------
bool CThread::isSuspended() 
{
    //If mutex locks, thread is suspended
    int retVal = pthread_mutex_trylock(&_mutex_suspend);
    if(0 == retVal)
    {   
        //Unlock mutex and return true for suspended state
        pthread_mutex_unlock(&_mutex_suspend);
        return true;
    }
    //Mutex was alread locked, thread working
    return false;
}

//--------------------------------------------------------------------------------------------
// Joins this thread 
//--------------------------------------------------------------------------------------------
bool CThread::join() 
{
#ifdef DEBUG
    printf("CThread %d: Joining thread.\n", _id);
#endif    

    //Lock mutex to make sure thread is sleeping
    pthread_mutex_lock(&_mutex_suspend);

    //Set thread to exit
    _exit = true;

#ifdef DEBUG
    printf("CThread %d Join(): Thread suspended. Wake it for exit.\n", _id);
#endif

    //Wake up thread
    int retVal = pthread_cond_signal(&_cond_resume);
    if(0 != retVal)
    {
        printf("CThread.cpp: Error resuming thread. Error: %d\n", retVal);
    }
    
    pthread_mutex_unlock(&_mutex_suspend);

    void *status;    
    retVal = pthread_join(_thread, &status);
    if (0 != retVal)
    {
        printf("CThread.cpp: ERROR; return code from pthread_join() is %d\n", retVal);
        return false;
    }                            
#ifdef DEBUG
    printf("CThread %d: Join status: %ld\n", _id, (long)status); 
#endif
    return true;
}

//--------------------------------------------------------------------------------------------
// The actual thread work
// Threads pop new tasks from queue or takes parent task until no tasks left
// Threads with no task suspends until resumed with a new task or exit is signaled
//--------------------------------------------------------------------------------------------
void* CThread::theWork(void* ptr) 
{
    //Get the pointer to self
    CThread* pThread = (CThread*) ptr;
    int retVal;

    //Lock the mutex
    retVal = pthread_mutex_lock(&pThread->_mutex_suspend);
    if(0 != retVal)
        printf("CThread.cpp: Error locking thread mutex. Error: %d\n", retVal);

    while(1)
    {
        //Check to see if is suspended (must set suspend mutex before call to cond_wait)
        if(true == pThread->_suspend && false == pThread->exit())
        {
#ifdef DEBUG
            printf("CThread %d: thread suspended\n", pThread->getId());
#endif
            //Put self in ready queue
            pThread->getThreadHandler()->pushReadyThread(pThread);

            //Suspend the thread and wait for resume
            retVal = pthread_cond_wait(&pThread->_cond_resume, &pThread->_mutex_suspend);            
            if(0 != retVal)
                printf("CThread.cpp: Error waiting for condition. Error: %d\n", retVal);
        }
        
        //Exit thread if set to destroy
        if(true == pThread->exit())
        {   
#ifdef DEBUG
            printf("CThread %d: exiting thread\n", pThread->getId());
#endif
            //Unlock mutex
            retVal = pthread_mutex_unlock(&pThread->_mutex_suspend);      

            //Thread destroyed
            pthread_exit(0);
        }
#ifdef DEBUG
        printf("CThread %d: Calling Work() with task %d\n", pThread->getId(), pThread->getTask()->getId());
#endif
        //Execute task
         pThread->getTask()->Work();

#ifdef DEBUG
        printf("CThread %d: Done Work() on task %d\n", pThread->getId(), pThread->getTask()->getId());
#endif
        //Set next task if ready 
        if(0 != pThread->getTask()->getParent() && 0 == pThread->getTask()->getParent()->decrChildCounter())
        {   
#ifdef DEBUG
            printf("*** CThread %d: Task finished last. Taking parent job!! ***\n", pThread->getId());
#endif
            pThread->setTask(pThread->getTask()->getParent());
        }
        else
        {
	    // Signal this root task's main thread to wake up
	    if (true == pThread->getTask()->isRootTask() && 0 != pThread->getTask()->getTaskQueue())
	    {
//		printf("CThread %d, Work: Root task %d signaling main thread\n", pThread->getId(), pThread->getTask()->getId());
		pThread->getTask()->getTaskQueue()->signalMainThread(); 
	    }

	    // Get a task queue from the thread handler
//	    TaskQueue* tmpQueue = pThread->getThreadHandler()->getQueue();
	    ThreadHandler* threadHandler = pThread->getThreadHandler();
	    CTask* task = 0;

	    // Get new task from queue
//	    if (0 != tmpQueue)
//	    {
//		task = tmpQueue->pop(); 
//	    }

	    if (0 != threadHandler)
	    {
		task = threadHandler->getTask(); 
	    }

	    //Suspend thread if no more work available
	    if(0 == task)
	    {
		pThread->_suspend = true;
		continue;
	    }

            //Set new task for thread
            pThread->setTask(task);
        }
        
        //Make thread skip suspend and directly check for new task or work on parent task
        pThread->_suspend = false;
    }
}

void CThread::print() const
{   
    printf("Thread_ID: %d, Current Task_Id: %d\n", _id, _task->getId());
}

void CThread::setTask(CTask* in)
{
//    //Delete old task form heap
//    if(0 != _task)
//        delete _task;

    _task = in;
}
