#include "TaskQueue.h"
#include "CTask.h"
#include <stdio.h>

//-------------------------------------------
// Ctor
//-------------------------------------------
TaskQueue::TaskQueue() :
    _hasInitThreadAttrs(false),
//    _rootTasks(0),
    _numRootTasks(0),
    _position(0),
    _doneDistributingWork(false)
{
    if(false == _hasInitThreadAttrs)
        initThreadAttributes();
}

//--------------------------------------------------------------
// Copy Ctor
//--------------------------------------------------------------
TaskQueue::TaskQueue(const TaskQueue & in) :
    _hasInitThreadAttrs( in._hasInitThreadAttrs ),
    _rootTasks( in._rootTasks ),
    _numRootTasks( in._numRootTasks),
    _position( in._position ),
    _doneDistributingWork( in._doneDistributingWork )
{
    if(false == _hasInitThreadAttrs)
        initThreadAttributes();
}

//--------------------------------------------------------------
// Equal operator
//--------------------------------------------------------------a
TaskQueue & TaskQueue::operator=(const TaskQueue & in)
{
    _hasInitThreadAttrs = in._hasInitThreadAttrs;
    _rootTasks = in._rootTasks;
    _numRootTasks = in._numRootTasks;
    _position = in._position;
    _doneDistributingWork = in._doneDistributingWork;

    if(false == _hasInitThreadAttrs)
        initThreadAttributes();
}

//-------------------------------------------
// Dtor
//-------------------------------------------
TaskQueue::~TaskQueue()
{
    pthread_mutex_lock(&_mutex);

    // Delete root task. Child tasks deleted recursively
    if (0 != _rootTasks.size())
    {
	for (int i = 0; i < _rootTasks.size(); i++)
	{
	    if (0 != _rootTasks[i])
	    {
		delete _rootTasks[i];		
	    }
	}
    }
    else if (0 != _queue.size())
    {
	//printf("Root task not set for queue! Memory might not be cleaned up!\n");
	for (int i = 0; i < _queue.size(); i++)
	{
	    if (0 != _queue[i])
	    {
		delete _queue[i];
		_queue[i] = 0; 
	    }
	}
	_queue.clear();
    }

    pthread_mutex_unlock(&_mutex);

    if(true == _hasInitThreadAttrs)
    {
        //Destroy mutexes
        pthread_mutex_destroy(&_mutex_root_tasks);
        pthread_mutex_destroy(&_mutex);
        pthread_mutex_destroy(&_mutex_main_thread_barrier);
        pthread_cond_destroy(&_cond_main_thread_barrier);
    }
}

//-------------------------------------------
// Initialize threading attributes
//-------------------------------------------
void TaskQueue::initThreadAttributes()
{
    pthread_mutex_init(&_mutex_root_tasks, NULL);
    pthread_mutex_init(&_mutex, NULL);
    pthread_mutex_init(&_mutex_main_thread_barrier, NULL);
    pthread_cond_init(&_cond_main_thread_barrier, NULL);
    _hasInitThreadAttrs = true;
}

//-------------------------------------------
// Suspend this queue's main thread
//-------------------------------------------
void TaskQueue::suspendMainThread()
{
    //Make main thread wait for all tasks to finish
    pthread_mutex_lock(&_mutex_main_thread_barrier);
    if(false == _doneDistributingWork)
    {
        _doneDistributingWork = true;
        pthread_cond_wait(&_cond_main_thread_barrier, &_mutex_main_thread_barrier);
    }
    pthread_mutex_unlock(&_mutex_main_thread_barrier);
}

//-------------------------------------------
// Signal this queue's main thread to continue
//-------------------------------------------
int TaskQueue::signalMainThread()
{
    int retVal = 0;
    pthread_mutex_lock(&_mutex_main_thread_barrier);

    // Check if all root tasks have finished
    retVal = decrementRootCounter();
    if (retVal <= 0)
    {	
	// Signal main thread
	if(true == _doneDistributingWork)
	{
	    pthread_cond_signal(&_cond_main_thread_barrier);
	}
	else
	{
	    _doneDistributingWork = true;
	}
    }
    
    pthread_mutex_unlock(&_mutex_main_thread_barrier);
    return retVal;
}

//-------------------------------------------
// Thread safe reset of queue position
//-------------------------------------------
void TaskQueue::resetQueue()
{
    pthread_mutex_lock(&_mutex);
    _position = 0;
    _doneDistributingWork = false;
    pthread_mutex_unlock(&_mutex);
}

//-------------------------------------------
// Adds a task to root task list
//-------------------------------------------
void TaskQueue::addRootTask(CTask* in)
{
    if (0 == in)
    {
	printf("Trying to add 0 pointer as root task\n");
	return ; 
    }
    in->setIsRootTask(true);
    in->setTaskQueue(this);
    _rootTasks.push_back(in); 
    incrementRootCounter();    
}

//-------------------------------------------
// Set root task for queue and set pointer
// in root task back to queue for signaling
//-------------------------------------------
//void TaskQueue::setRootTask(CTask* task)
//{
//    if (0 == task)
//        return ;
//
//    _rootTask = task;
//    task->setIsRootTask(true);
//    task->setTaskQueue(this);
//}

//-------------------------------------------
// Decrement root node count and return result
//-------------------------------------------
int TaskQueue::decrementRootCounter()
{
    int retVal = 0;

    //Lock update mutex
    pthread_mutex_lock(&_mutex_root_tasks);
    
    //Save decremented counter in local variable
    retVal = --_numRootTasks;

    //Unlock update mutex
    pthread_mutex_unlock(&_mutex_root_tasks);
    return retVal; 
}

//-------------------------------------------
// Thread safe push of task in queue
//-------------------------------------------
void TaskQueue::push(CTask* task)
{
    //Lock mutex and add task to queue
    pthread_mutex_lock(&_mutex);
    _queue.push_back(task);
    pthread_mutex_unlock(&_mutex);
}

//-------------------------------------------
// Thread safe pop of task in queue
//-------------------------------------------
CTask* TaskQueue::pop()
{
    CTask* ptr = 0;
    //Lock mutex and pop task from queue
    pthread_mutex_lock(&_mutex);
    if (_position >= 0 && _position < _queue.size())
    {
        ptr = _queue[_position];
        _position++;
    }
    pthread_mutex_unlock(&_mutex);
    return ptr;
}

//-------------------------------------------
// Thread safe check if task queue is empty
//-------------------------------------------
bool TaskQueue::empty()
{
    bool empty = false;

    //Lock mutex and check queue size
    pthread_mutex_lock(&_mutex);
    if (_position >= _queue.size())
    {
        empty = true;
    }
    pthread_mutex_unlock(&_mutex);
    return empty;
}

//-------------------------------------------
// Serial traversial of the queue
//-------------------------------------------
void TaskQueue::traverseQueue()
{
    // Get initial task
    CTask* task = pop();
    while (true)
    {
	pthread_mutex_lock(&_mutex_main_thread_barrier);	
	bool done = _doneDistributingWork;
	pthread_mutex_unlock(&_mutex_main_thread_barrier);

        // Return if no task is received
        if (0 == task || true == done)
        {
	    // Done distributing work if there is no root task 
	    // (All tasks are in the queue, no parent pointers)
	    if(0 == getRootTasks().size())
	    {
		signalMainThread();
	    }
            return;
        }

        // Do the task
        task->Work();

	// Return if this was last root task
	if (true == task->isRootTask())
	{
	    if(signalMainThread() <= 0) 
		return;
	}

	// Get new task
        if(0 != task->getParent() && 0 == task->getParent()->decrChildCounter())
        {
	    // Set next task to parent if it's ready
            task = task->getParent();
        }
        else
        {
            // Get a new task
            task = pop();
        }
    }
}
