#include "ThreadHandler.h"
#include "CThread.h"
#include "CTask.h"
#include "TaskQueue.h"
#include "ThreadTimer.h"

int g_numThreads = 0;
ThreadHandler* ThreadHandler::_instance = 0;

//--------------------------------------------------------------
// Create a single instance of the thread handler
//--------------------------------------------------------------
ThreadHandler* ThreadHandler::Instance()
{
    if (0 == _instance)
    {
        _instance = new ThreadHandler();
    }
    return _instance;
}

//--------------------------------------------------------------
// Ctor
//--------------------------------------------------------------
ThreadHandler::ThreadHandler() :
    _hasInitAttributes(false),
    _doneDistributingWork(true)
{
    if(false == _hasInitAttributes)
    {
        initAttributes();
    }
}

//--------------------------------------------------------------
// Copy Ctor
//--------------------------------------------------------------
ThreadHandler::ThreadHandler(const ThreadHandler & in) :
    _hasInitAttributes(in._hasInitAttributes),
    _doneDistributingWork(in._doneDistributingWork)
{
    if(false == _hasInitAttributes)
    {
        initAttributes();
    }
}

//--------------------------------------------------------------
// Equal operator
//--------------------------------------------------------------
ThreadHandler & ThreadHandler::operator=(const ThreadHandler & in)
{
    if(this == &in) return *this;
    _hasInitAttributes = in._hasInitAttributes;
    _doneDistributingWork = in._doneDistributingWork;

    if(false == _hasInitAttributes)
    {
        initAttributes();
    }
    return *this;
}

//--------------------------------------------------------------
// Dtor
//--------------------------------------------------------------
ThreadHandler::~ThreadHandler()
{
    // Join all threads
    joinThreads();

    //Delete threads
    for(size_t i = 0; i < _threadPool.size(); i++)
    {
	if(0 != _threadPool[i])
	    delete _threadPool[i];
    }
    _threadPool.clear();

    pthread_mutex_destroy(&_mutex);
    pthread_mutex_destroy(&_mutex_task_queues);
    pthread_mutex_destroy(&_mutex_main_thread_barrier);
    pthread_cond_destroy(&_cond_main_thread_barrier);
}

//--------------------------------------------------------------
// Adds a task queue to the list of task queues
//--------------------------------------------------------------
void ThreadHandler::addTaskQueue(TaskQueue* in)
{
    if(0 != in)
    {
        pthread_mutex_lock(&_mutex_task_queues);

        // Reset queue position
        in->resetQueue();

        // Add queue til queue stack
        _taskQueues.push(in);

        pthread_mutex_unlock(&_mutex_task_queues);
    }
    else
    {
        printf("Task queue is NULL!\n");
    }
}

//--------------------------------------------------------------
// Adds num threads to the thread pool
//--------------------------------------------------------------
void ThreadHandler::addThreads(const int num)
{
#ifdef DEBUG
    printf("ThreadHandler: Adding threads\n");
#endif

    if(0 == num)
    {
	printf("Trying to add 0 threads. Must at least add 1 thread\n");
	return;
    }

    //Lock main thread barrier while creating threads
    pthread_mutex_lock(&_mutex_main_thread_barrier);

    //Add threads to pool
    for(size_t i = 0; i < num; i++)
    {
        CThread* thread = new CThread();
        thread->setId(i);
        thread->setThreadHandler(this);
        _threadPool.push_back(thread);
    }
    
    // Set global number of threads
    g_numThreads = _threadPool.size();

    //Create available threads
    //Must create threads here for threadPool/readyThreadPool size correctness
    for(size_t i = 0; i < num; i++)
    {
        _threadPool[i]->create();
    }

    //Wait until all threads are in suspend mode
    //Last ready thread will signal this condition
    pthread_cond_wait(&_cond_main_thread_barrier, &_mutex_main_thread_barrier);
    pthread_mutex_unlock(&_mutex_main_thread_barrier);

#ifdef DEBUG
    printf("ThreadHandler: All threads ready\n");
#endif

}

//--------------------------------------------------------------
// Push a suspended thread on the ready thread stack
// Used by the worker threads
//--------------------------------------------------------------
void ThreadHandler::pushReadyThread(CThread* thread)
{
    //Lock ready thread stack access
    pthread_mutex_lock(&_mutex);

    //Push thread on ready thread stack
    _readyThreadPool.push(thread);

    //Wake main thread if all threads are back in ready pool
    if(_readyThreadPool.size() == _threadPool.size())
    {
        pthread_mutex_lock(&_mutex_main_thread_barrier);
        if(true == _doneDistributingWork)
        {
            pthread_cond_signal(&_cond_main_thread_barrier);
        }
        else
        {
            _doneDistributingWork = true;
        }
        pthread_mutex_unlock(&_mutex_main_thread_barrier);
    }
    pthread_mutex_unlock(&_mutex);
}

//--------------------------------------------------------------
// Kills num threads from the thread pool that are in the
// ready thread pool
//--------------------------------------------------------------
void ThreadHandler::killThreads(int numToKill)
{
    // Do not kill more than there are threads
    numToKill = (numToKill >= _threadPool.size()) ? _threadPool.size() : numToKill;

    //Lock ready thread stack access
    pthread_mutex_lock(&_mutex);

    int counter = 0;
    while (counter <= numToKill && false == _readyThreadPool.empty())
    {
        CThread* ptr = 0;
	CThread* ptr2 = 0;

	//Get pointer to a ready thread
	ptr = _readyThreadPool.top();

	// Find same thread in taskpool
	for(int i = _threadPool.size() - 1; i >= 0; i--)
	{
	    ptr2 = _threadPool[i];
		
	    if(0 == ptr || 0 == ptr2)
	    {
		printf("Error killing threads!\n");
		continue;
	    }
	    else if(ptr->getId() == ptr2->getId())
	    {
		// Delete the thread from the pool and clean up memory
		ptr2->join();
		_threadPool.erase(_threadPool.begin() + i);
		delete ptr2;
	    }
	    
	    // Pop the thread from the ready queue as well
            _readyThreadPool.pop();
        }
        counter++;
    }

    // Set global thread number
    g_numThreads = _threadPool.size();

    printf("Killed %d threads!\n", counter);

    pthread_mutex_unlock(&_mutex);
}

//--------------------------------------------------------------
// Pops a suspended thread from the ready thread stack
// Used by the main thread
//--------------------------------------------------------------
CThread* ThreadHandler::popReadyThread()
{
    //Lock ready thread stack access
    pthread_mutex_lock(&_mutex);

    CThread* ptr = 0;
    if(false == _readyThreadPool.empty())
    {
        //Get pointer to a ready thread
        ptr = _readyThreadPool.top();
        _readyThreadPool.pop();
    }
    pthread_mutex_unlock(&_mutex);
    return ptr;
}

//--------------------------------------------------------------
// Checks if the ready thread stack is empty
//--------------------------------------------------------------
bool ThreadHandler::readyThreadPoolEmpty()
{
    pthread_mutex_lock(&_mutex);
    bool empty = _readyThreadPool.empty();
    pthread_mutex_unlock(&_mutex);
    return empty;
}

//--------------------------------------------------------------
// Initializes the threading attributes
//--------------------------------------------------------------
void ThreadHandler::initAttributes()
{
    //Init attributes
    _hasInitAttributes = true;
    pthread_mutex_init(&_mutex, NULL);
    pthread_mutex_init(&_mutex_task_queues, NULL);
    pthread_mutex_init(&_mutex_main_thread_barrier, NULL);
    pthread_cond_init(&_cond_main_thread_barrier, NULL);
}

//--------------------------------------------------------------
// Distributes work to all threads on the ready thread stack
// Main thread suspends after work is distributed
// Worker threads pops new tasks directly
// Last worker thread signals main thread to wake and continue
// Used by the main thread
//--------------------------------------------------------------
void ThreadHandler::distributeWork(TaskQueue* queue)
{
#ifdef DEBUG
    printf("ThreadHandler: Distributing work\n");
#endif
    if(0 == queue)
        return;

    // Reset queue position
    queue->resetQueue();

//    printf("Queue size %d\n", queue->getQueue().size());

    pthread_mutex_lock(&_mutex_task_queues);

    // Add queue to queue stack
    _taskQueues.push(queue);

    pthread_mutex_unlock(&_mutex_task_queues);

    int readyThreadsFound = 0;

    //Loop until all workers have work or queue is empty
    while(false == readyThreadPoolEmpty())
    {
        //Get work from queue
//        CTask* task = queue->pop();
	CTask* task = getTask();

        if (0 == task)
        {
            break;
        }

#ifdef DEBUG
        if(0 != task)
            printf("ThreadHandler: Task %d popped from queue\n", task->getId());
#endif
        //Resume next available thread
        CThread* thread = popReadyThread();

        if(0 != thread)
        {
#ifdef DEBUG
            printf("ThreadHandler: Resuming worker %d with task %d\n", thread->getId(), task->getId());
#endif
            //Set thread task and resume thread
            thread->resume(task);

	    readyThreadsFound++;
        }
        else
        {
            printf("ThreadHandler: ERROR: No task or thread received!\n");
        }
    }

    // Make main thread join work for this queue
    queue->traverseQueue();

    // Make main thread wait for it's queue's tasks to finish
//    printf("Suspending. Distributed tasks to %d threads\n", readyThreadsFound);
    queue->suspendMainThread();
//    printf("Continuing\n");
}

//--------------------------------------------------------------
// Join all threads
//--------------------------------------------------------------
void ThreadHandler::joinThreads()
{
    //Join all threads
    for(size_t i = 0; i < _threadPool.size(); i++)
    {
        _threadPool[i]->join();
    }
}

//--------------------------------------------------------------
// Finds a queue with tasks and pops empty queues off the queue
// stack. Also signals popped queue's main thread to wake up if
// the tasks in the queue have no 'hidden' parent tasks. If there
// are parent tasks, the thread that finishes the root task
// sends the signal
//--------------------------------------------------------------
TaskQueue * ThreadHandler::getQueue()
{
    pthread_mutex_lock(&_mutex_task_queues);
    TaskQueue* ptr = (false == _taskQueues.empty()) ? _taskQueues.top() : 0;
    while (0 != ptr && true == ptr->empty())
    {
        // If there are no parent tasks, signal this queue's main thread to continue
        if (false == ptr->hasRootTasks())
        {
            ptr->signalMainThread();
        }
        _taskQueues.pop();
        ptr = (false == _taskQueues.empty()) ? _taskQueues.top() : 0;
    }
    pthread_mutex_unlock(&_mutex_task_queues);

    return ptr;
}

//--------------------------------------------------------------
// Finds a queue with tasks and pops empty queues off the queue
// stack. Also signals popped queue's main thread to wake up if
// the tasks in the queue have no 'hidden' parent tasks. If there
// are parent tasks, the thread that finishes the root task
// sends the signal
//
// Finally it pops a task from the queue and returns it
//--------------------------------------------------------------
CTask * ThreadHandler::getTask()
{
    pthread_mutex_lock(&_mutex_task_queues);

    // Get queue on top of stack
    TaskQueue* ptr = (false == _taskQueues.empty()) ? _taskQueues.top() : 0;
    while (0 != ptr && true == ptr->empty())
    {
        // If there are no parent tasks, signal this queue's main thread to continue
        if (false == ptr->hasRootTasks())
        {
            ptr->signalMainThread();
        }
        _taskQueues.pop();
        ptr = (false == _taskQueues.empty()) ? _taskQueues.top() : 0;
    }

    // Get task from the found queue
    CTask* tmpTask = 0;
    if (0 != ptr)
    {
	tmpTask = ptr->pop();
    }

    pthread_mutex_unlock(&_mutex_task_queues);

    return tmpTask;
}

//--------------------------------------------------------------
// Alters the number of threads if the specified number of 
// timings have been saved
//--------------------------------------------------------------
void ThreadHandler::addTiming(int & index, const double time)
{
    ThreadTimer::Instance()->get_thread_timer(index)->save_time(getNumThreads(), time);
    _timerCounter++;

    if (0 == _timerCounter % ThreadTimer::Instance()->get_num_timings())
    {
	// Reset the counter
	_timerCounter = 0;
	
	// Double number of threads
	addThreads(getNumThreads());
    }

//    // TODO: Get num CPU's dynamically (stop on 2xNumCPUs)
//    if (getNumThreads() => 8)
//    {
//	
//    }
}
